/*****************************************************************************
 * swapfields.c : swapfields plugin for vlc
 *****************************************************************************
 * Copyright (C) 2003-2007 the VideoLAN team
 * $Id: invert.c 18062 2006-11-26 14:20:34Z zorglub $
 *
 * Author: Paul Corke <paulc at datatote dot co dot uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Description
 *****************************************************************************
 * This filter is useful when transcoding from a PAL DV source to an
 * interlaced format and subsequently displaying it with a deinterlacer such
 * as "bob" or "linear".  PAL DV has the field order backwards which causes
 * the deinterlacer to display the second frame before the first which leads
 * to a juddering of the picture.  This filter moves the entire image up by
 * one row which corrects the field order.
 *
 * There is an optional smoothing that can be used to reduce the flicker that
 * occurs with bob deinterlacing however this results in some motion blur on
 * the resulting image.
 *
 * Sample use:
 *
 * dvgrab - | vlc - --sout'#transcode{vcodec=mp2v,vb=6000,acodec=mp2a,ab=192,\
 * vfilter=swapfields{deflicker=1}}:standard{access=file,mux=ps,dst=video.ps}'
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include <stdlib.h>                                      /* malloc(), free() */
#include <string.h>

#include <vlc/vlc.h>
#include <vlc_vout.h>

#include "vlc_filter.h"

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static int  Create      ( vlc_object_t * );
static void Destroy     ( vlc_object_t * );

static picture_t *Filter( filter_t *, picture_t * );
static int DeflickerCallback( vlc_object_t *, char const *,
                            vlc_value_t, vlc_value_t, void * );

#define FILTER_PREFIX "swapfields-"

#define DEFLICKER_TEXT N_( "Deflicker image 0,1,2." )
#define DEFLICKER_LONGTEXT N_( "Set the deflicker strength, 0 is off, 1 is light, 2 is strong.  Reduces flicker when deinterlacing later." )


/*****************************************************************************
 * Module descriptor
 *****************************************************************************/
vlc_module_begin();
    set_description( _("Swap Fields video filter") );
    set_shortname( _("Swap Fields" ));
    set_category( CAT_VIDEO );
    set_subcategory( SUBCAT_VIDEO_VFILTER );
    set_capability( "video filter2", 0 );
    add_integer_with_range( FILTER_PREFIX "deflicker", 0, 0, 2, NULL,
        DEFLICKER_TEXT, DEFLICKER_LONGTEXT, VLC_FALSE );
    add_shortcut( "swapfields" );
    set_callbacks( Create, Destroy );
vlc_module_end();

static const char *ppsz_filter_options[] = {
    "deflicker", NULL
};

/*****************************************************************************
 * vout_sys_t: Swap Fields video output method descriptor
 *****************************************************************************
 * This structure is part of the video output thread descriptor.
 * It describes the Invert specific properties of an output thread.
 *****************************************************************************/
struct filter_sys_t
{
    uint8_t i_deflicker;
};

/*****************************************************************************
 * Create: allocates Swap Fields video thread output method
 *****************************************************************************
 * This function allocates and initializes a Invert vout method.
 *****************************************************************************/
static int Create( vlc_object_t *p_this )
{
    filter_t *p_filter = (filter_t *)p_this;

    /* Allocate structure */
    p_filter->p_sys = malloc( sizeof( filter_sys_t ) );
    if( p_filter->p_sys == NULL )
    {
        msg_Err( p_filter, "out of memory" );
        return VLC_ENOMEM;
    }

    p_filter->pf_video_filter = Filter;

    config_ChainParse( p_filter, FILTER_PREFIX, ppsz_filter_options,
                   p_filter->p_cfg );

    p_filter->p_sys->i_deflicker =
        var_CreateGetIntegerCommand( p_filter, FILTER_PREFIX "deflicker" );
    var_AddCallback( p_filter, FILTER_PREFIX "deflicker",
                     DeflickerCallback, p_filter->p_sys );

    return VLC_SUCCESS;
}

/*****************************************************************************
 * Destroy: destroy Swap Fields video thread output method
 *****************************************************************************
 * Terminate an output method created by Create()
 *****************************************************************************/
static void Destroy( vlc_object_t *p_this )
{
    filter_t *p_filter = (filter_t *)p_this;
    free( p_filter->p_sys );
}

/*****************************************************************************
 * Render: displays previously rendered output
 *****************************************************************************
 * This function changes the field order of the specified picture
 *****************************************************************************/
static picture_t *Filter( filter_t *p_filter, picture_t *p_pic )
{
    int i, j;
    uint8_t *p_src = NULL;
    uint8_t *p_out = NULL;
    int i_src_pitch;

    if( !p_pic ) return NULL;
    if( !p_filter ) return NULL;

    /* process the Y plane */
    p_src = p_out = p_pic->p[Y_PLANE].p_pixels;
    if( !p_src )
    {
        msg_Warn( p_filter, "Can't get Y plane." );
        if( p_pic->pf_release )
            p_pic->pf_release( p_pic );
        return NULL;
    }

    i_src_pitch = p_pic->p[Y_PLANE].i_visible_pitch;

    /* Shift everything on Y plane up one row. */
    for( i = 0; i < ( p_pic->p[Y_PLANE].i_visible_lines - 1 ); i++ )
    {
        switch( p_filter->p_sys->i_deflicker )
        {
            case 1:
            {
                /* Low de-flicker.  3x lower field, 1x current field */
                for( j = 0; j < p_pic->p[Y_PLANE].i_pitch; ++j)
                {
                    int p = p_src[p_pic->p[Y_PLANE].i_pitch];
                    p = p + ( p << 1 ) + ( *p_src++ ); 
                    *p_out++ = ( p >> 2 );
                }
                break;
            }
            case 2:
            {
                /* Strong de-flicker.  2x lower field, 1x current field */
                for( j = 0; j < p_pic->p[Y_PLANE].i_pitch; ++j)
                {
                    int p = p_src[p_pic->p[Y_PLANE].i_pitch];
                    p = ( p << 1 ) + ( *p_src++ );
                    *p_out++ = ( p / 3 );
                }
                break;
            }

            default:
            {
                /* No de-flicker.  1x lower field (Just copy everything up). */
                p_filter->p_libvlc->pf_memcpy( p_out,
                                               p_src + p_pic->p[Y_PLANE].i_pitch,
                                               p_pic->p[Y_PLANE].i_pitch );
                p_src += p_pic->p[Y_PLANE].i_pitch;
                p_out += p_pic->p[Y_PLANE].i_pitch;
                break;
            }
        }
    }

    return p_pic;
}

static int DeflickerCallback( vlc_object_t *p_this, char const *psz_var,
                            vlc_value_t oldval, vlc_value_t newval,
                            void *p_data )
{
    filter_sys_t *p_sys = (filter_sys_t *)p_data;
    if( !strcmp( psz_var, FILTER_PREFIX "deflicker" ) )
        p_sys->i_deflicker = __MIN( 2, __MAX( 0, newval.i_int ) );
    return VLC_SUCCESS;
}

